@@ -25,6 +25,21 @@
 
 ;; Comment:
 
+;;;
+;;; Character representation.
+;;;    CCL can't treat character itself as it is, as well as Emacs.
+;;;    Since CCL can deal with only numbers currently, we must manage
+;;;    to convert character to number.
+;;;
+;;;    But, like XEmacs, some systems distinguish between number and
+;;;    character.  In addition, Emacs may give some abstraction
+;;;    to characters.  Therefore MULE-UCS should prepare certain
+;;;    abstraction layer to represent character format on CCL.
+
+;;;    MULE-UCS prepare `TYPE' to give any kind of representations
+;;;    defined by users to data sturucture to be translated.
+;;;    Users can define any representation on translation in their fashion.
+
 ;;; TYPE specification.
 ;;;      TYPE on MULE-UCS is for specifying character data categories that
 ;;; are distinguished one another whenever we treat it in CCL environment
@@ -60,15 +75,19 @@
 
 (defvar mucs-type-alist nil)
 
+(defsubst mucs-type-p (type)
+  (and (assq type mucs-type-alist)
+       t))
+
 (defsubst mucs-type-get-option-alist (type)
   (nth 4 (or (assq type mucs-type-alist)
-	     (error "TYPE:%S is not registered" type))))
+	     (signal 'mucs-unknown-type-error type))))
 
 (defsubst mucs-type-set-option-alist (type alist)
   (let ((slot (assq type mucs-type-alist)))
     (if slot
 	(setcar (nthcdr 4 slot) alist)
-      (error "TYPE:%S is not registered" type))))
+      (signal 'mucs-unknown-type-error type))))
 
 (defsubst mucs-type-assq-option-alist (type key)
   (assq key (mucs-type-get-option-alist type)))
@@ -101,12 +120,21 @@
 (defun mucs-define-type (type elisp-conv ccl-conv
 			      &optional equal-op relative-op)
   "Define TYPE."
+
+  (if (null equal-op)
+      (setq equal-op (function eq)))
+  (if (null relative-op)
+      (setq relative-op (function <)))
+
   ;; check conv
   (mucs-representation-check-conversion-function elisp-conv)
   (mucs-representation-check-conversion-function ccl-conv)
   (let ((slot (assq type
 		    mucs-type-alist))
-	(cell (list type elisp-conv ccl-conv nil nil)))
+	(cell (list type
+		    (cons elisp-conv ccl-conv)
+		    (cons equal-op relative-op)
+		    nil nil)))
     (if slot
 	(progn
 	  (setcdr slot (cdr cell))
@@ -123,15 +151,30 @@
   (let ((slot (assq type
 		    mucs-type-alist)))
     (if (null slot)
-	(error "TYPE:%S is not registered" type)
-      (nth 1 slot))))
+	(signal 'mucs-unknown-type-error type)
+      (car (nth 1 slot)))))
 
 (defun mucs-type-get-ccl-representation (type)
   (let ((slot (assq type
 		    mucs-type-alist)))
     (if (null slot)
-	(error "TYPE:%S is not registered" type)
-      (nth 2 slot))))
+	(signal 'mucs-unknown-type-error type)
+      (cdr (nth 1 slot)))))
+
+(defun mucs-type-get-equal-op (type)
+  (let ((slot (assq type
+		    mucs-type-alist)))
+    (if (null slot)
+	(signal 'mucs-unknown-type-error type)
+      (car (nth 2 slot)))))
+
+(defun mucs-type-get-relative-op (type)
+  (let ((slot (assq type
+		    mucs-type-alist)))
+    (if (null slot)
+	(signal 'mucs-unknown-type-error type)
+      (cdr (nth 2 slot)))))
+
 
 ;;
 ;; Serialization
@@ -140,7 +183,7 @@
 (defsubst mucs-type-check-valid-serialization (serialize)
   (if (not (or (listp serialize)
 	       (eq 'none serialize)))
-      (error "Invalid serialization:%S" serialize)))
+      (signal 'mucs-invalid-serialization-error serialize)))
 
 (defun mucs-type-register-serialization (type name serialize unserialize)
   (if (not (symbolp name))
@@ -154,7 +197,7 @@
 	serial-slot
 	serials)
     (if (null slot)
-	(error "TYPE:%S is not defined." type))
+	(signal 'mucs-unknown-type-error type))
     (setq serials (nth 3 slot)
 	  serial-slot (assq name serials)
 	  cell (cons name (cons serialize unserialize)))
@@ -171,16 +214,18 @@
 	       (nth 3 (assq type mucs-type-alist))))
 	result)
     (if (null serialize-slot)
-	(error "have not defined serialization-method %S of %S"
-	       name type))
+	(signal 'mucs-unknown-serialization-error
+		(cons name type)))
     (setq result
 	  (if unserializep
 	      (cdr (cdr serialize-slot))
 	    (car (cdr serialize-slot))))
     (if (eq result 'none)
 	(if unserializep
-	    (error "TYPE:%S has no unserialization method of %S" type name)
-	  (error "TYPE:%S has no serialization method of %S" type name))
+	    (signal 'mucs-no-serialization-method
+		    (cons type name))
+	  (signal 'mucs-no-unserialization-method
+		  (cons type name)))
       result)))
 
 
@@ -193,7 +238,7 @@
 	to-type-conv-slot
 	conv-func-slot)
     (if (null slot)
-	(error "TYPE:%S is not defined." from-type))
+	(signal 'mucs-unknown-type-error from-type))
     (mucs-representation-check-conversion-function conversion)
 
     (setq to-type-conv-slot (mucs-type-assq-option-alist
@@ -217,7 +262,7 @@
 	to-type-conv-slot
 	conv-func-slot)
     (if (null slot)
-	(error "TYPE:%S is not defined." from-type))
+	(signal 'mucs-unknown-type-error from-type))
 
     (setq to-type-conv-slot (mucs-type-assq-option-alist
 			     from-type 'type-conversion)
@@ -237,7 +282,7 @@
 (defun mucs-type-serialization-list (type)
   (let ((slot (assq type mucs-type-alist)))
     (if (null slot)
-	(error "TYPE:%S is not defined." type))
+	(signal 'mucs-unknown-type-error type))
     (mapcar
      'car
      (nth 3 slot))))
