@@ -1,5 +1,5 @@
 ;;; -*- coding: iso-2022-7bit  -*-
-;;; tae.el --- Translate And Encoding compiler(TAE:$BL/(B):-)
+;;; tae.el --- Translation rule And Encoding compiler(TAE:$BL/(B):-)
 
 ;; Copyright (C) 1997-1999 Miyashita Hisashi
 
@@ -27,13 +27,15 @@
 ;;; TAE is CORE module. But this is VERY insufficient version.
 
 (require 'mucs)
-(require 'mucs-ccl)
 (require 'mucs-type)
 (require 'tbl-mg)
 
 
 (defun tae-project-all-translation-vector (val)
-  (vector t val 0 (mucs-max-code)))
+  (vector t
+	  (if (eq val 'invalid)
+	      mucs-invalid-code val)
+	  0 (mucs-max-code)))
 
 ;(register-code-conversion-map
 ; 'tae-cease-translation-table
@@ -41,7 +43,8 @@
 
 
 (defvar tae-translation-table-set-index 1)
-(defsubst tae-generate-translation-table-set-name (tr-name &optional decodep index)
+(defsubst tae-generate-translation-table-set-name
+  (tr-name &optional decodep index)
     "Return a name of a symbol used to register a table-set,
 which is managed by TAE."
     (if (null index)
@@ -210,7 +213,7 @@ after its compilation."
 ;;;;;;   that matches anything.  For example, an A-LIST, ((1 . 5) (all . 10))
 ;;;;;;   directs to translate 1 to 5, and any other number to 10 in encoding.
 ;;;;;;   In decoding, TAE ignores the (all . 10) element, because it is
-;;;;;;   impossible translate a number to everything.
+;;;;;;   impossible to translate a number to everything.
 ;;;;;;     And, in another special case, it can also have 'invalid(SYMBOL)
 ;;;;;;   that matches nothing and if a translated result is invalid, it
 ;;;;;;   ceases any further translation, and make the CCL state invalid.
@@ -831,14 +834,15 @@ according to it, rearrange the elements of UNIT."
 	(alist2 mes2)
 	elem1 elem2)
     (while (setq elem1 (car alist1))
-      (if (memq elem1 '(new-type))     ;;;; not merged key list.
-	  nil
-	(setq elem2 (assq (car elem1) alist2)
-	      alist1 (cdr alist1))
-	(if elem2
-	    (progn
+      (setq elem2 (assq (car elem1) alist2)
+	    alist1 (cdr alist1))
+      (if elem2
+	  (progn
+	    (setq alist2 (delq elem2 alist2))
+	    (if (memq (car elem1) '(new-type))     ;;;; non merged key list.
+		(setcdr elem1 (cdr elem2))
 	      (nconc elem1 (cdr elem2))
-	      (setq alist2 (delq elem2 alist2))))))
+	      ))))
     (nconc mes1 alist2)))
 
 (defsubst tae-message-add (message key data)
@@ -851,7 +855,7 @@ according to it, rearrange the elements of UNIT."
 (defsubst tae-message-put (message key value)
   (setcdr (assq key message) value))
 
-(defsubst tae-message-table-add (message &rest tables)
+(defsubst tae-message-add-tables (message tables)
   (let ((slot (assq 'tables message)))
     (if slot
 	(progn
@@ -860,7 +864,15 @@ according to it, rearrange the elements of UNIT."
 	  message)
       (cons (cons 'tables tables) message))))
 
-(defsubst tae-message-new-type-set (message type)
+(defsubst tae-message-set-table-set (message table-set)
+  (let ((slot (assq 'table-set message)))
+    (if slot
+	(progn
+	  (setcdr slot type)
+	  message)
+      (cons (cons 'new-type type) message))))
+
+(defsubst tae-message-set-new-type (message type)
   (let ((slot (assq 'new-type message)))
     (if slot
 	(progn
@@ -868,6 +880,14 @@ according to it, rearrange the elements of UNIT."
 	  message)
       (cons (cons 'new-type type) message))))
 
+(defun tae-generate-new-table-set (tables name decodep)
+  (let ((table-set
+	 (intern
+	  (tae-generate-translation-table-set-name name decodep))))
+    (tae-translation-add-table-set name table-set)
+    (define-table-set table-set tables)
+    table-set))
+
 (defun tae-compile-translation-element (tr-elem type decodep)
   (let* ((type-spec (tae-get-tr-elem-type-spec tr-elem))
 	 (from-type (if decodep
@@ -912,13 +932,11 @@ according to it, rearrange the elements of UNIT."
 			 (make-code-conversion-tables
 			  (tae-get-tr-elem-assoc tr-elem)
 			  conv-slot)
-			 ;;;;; We should replace nested table-set.???
 			 (if tr-elem-all-key
 			     (list
 			      (tae-project-all-translation-vector
 			       tr-elem-all-key))
-			   nil)
-			 ))))
+			   nil)))))
 
 		((tae-tr-elem-ccl-p tr-elem)
 		 (list (cons 'ccl (tae-get-tr-elem-ccl tr-elem))))
@@ -927,37 +945,39 @@ according to it, rearrange the elements of UNIT."
 		 (error "TAE cannot compile Elisp translation element directly:%S"
 			tr-elem))))
 
-    (tae-message-new-type-set result to-type)))
+    (tae-message-set-new-type result to-type)))
 
 (defun tae-compile-OR-operation (translations type decodep)
-  (let (tr result results to-type new-type)
+  (let (tr result results to-type new-type
+	   table-set-new-slot tables-new-slot)
     (while (setq tr (car translations))
       (setq translations (cdr translations)
-	    result (tae-get-compiled-products-internal tr type decodep)
+	    result (copy-sequence
+		    (tae-get-compiled-products-internal tr type decodep))
 	    new-type (cdr (assq 'new-type result)))
 
-      ;;; type operation.
+      ;;; operate type.
       (if to-type
 	  (if (not (eq to-type new-type))
-	      (error "TYPE mismatch:between %S and %S"
-		     to-type new-type))
+	      (signal 'mucs-type-mismatch-error to-type new-type))
 	(setq to-type new-type)
 	(setq results
-	      (tae-message-new-type-set results to-type)))
+	      (tae-message-set-new-type results to-type)))
+      ;;; operate 'table-set key.
+      (if (setq table-set-new-slot (assq 'table-set result))
+	  (progn
+	    (setq results (tae-message-add-tables
+			   results (list (cdr table-set-new-slot))))
+	    (setq result (delq table-set-new-slot result))))
+      ;;; operate 'tables key.
+      (if (setq tables-new-slot (assq 'tables result))
+	  (progn
+	    (setq results (tae-message-add-tables
+			   results (cdr tables-new-slot)))
+	    (setq result (delq tables-new-slot result))))
 
-      (setq results
-	    ;;;; only append .... :-P think more.
-	    ;;;; At this stage, tables of declared translation are
-	    ;;;; made into one (or more) table-set.  Hence, the tables
-	    ;;;; slot are NOT merged.  But we can retrieve
-	    ;;;; an alist like (... (table-set ...) ... (tables ...))
-	    ;;;; Since the tables slot will be made into a table-set slot at
-	    ;;;; later stage, we will have an alist like
-	    ;;;; (... (table-set) ... (table-set) ).
-	    ;;;; This is very dirty, thus I should implement
-	    ;;;; nested table-set as soon as possible.
-	    (tae-message-append results result))
-      )
+      ;; any other message keys are preserved.
+      (setq results (tae-message-append results result)))
     results))
 
 (defun tae-compile-AND-operation (translations type decodep)
@@ -991,11 +1011,7 @@ according to it, rearrange the elements of UNIT."
 (defun tae-get-compiled-products (name type decodep)
   (let ((products (tae-get-translation-produced-products
 		   name mucs-current-type decodep))
-	tables-slot
-	tables
-	table-set
-	reduced-sym
-	translation)
+	tables-slot reduced-sym	translation)
     (if products
 	products
       (setq reduced-sym (if decodep
@@ -1009,25 +1025,19 @@ according to it, rearrange the elements of UNIT."
 	    products (tae-get-compiled-products-internal
 		      translation type decodep))
 
-      ;;; convert tables to table-set.
-      (setq tables-slot (assq 'tables products))
-      (if tables-slot
+      ;; deal with 'tables message.
+      ;; convert it to table-set.
+      (if (setq tables-slot (assq 'tables products))
 	  (progn
-	    (setq tables (cdr tables-slot)
-		  table-set (intern
-			     (tae-generate-translation-table-set-name
-			      name decodep)))
-	    (if (null tables)
-		(error "In one TAE message, tables element exists more than once.")
-
-	      (tae-translation-add-table-set name table-set)
-	      (define-table-set table-set tables)
-
-	      (setcar tables-slot 'table-set)
-	      (setcdr tables-slot table-set))))
-
+	    (setcar tables-slot 'table-set)
+	    (setcdr tables-slot
+		    (tae-generate-new-table-set
+		     (cdr tables-slot) name decodep))))
+      ;; register products.
+      (setq foo products)
       (tae-set-translation-produced-products
        name type decodep products)
+      ;; return products.
       products)))
 
 ;; API for compilation.
@@ -1040,30 +1050,34 @@ NAME must be a declared translation."
       (error "Undeclared or invalid translation:%S" name))
   (let* ((type (or applied-type
 		   mucs-current-type))
-	 (products (tae-get-compiled-products name type decodep))
+	 (messages (tae-get-compiled-products name type decodep))
+	 (mes-cur messages)
 	 (result-myo (mucs-ccl-empty-myo))
 	 (tae-translation-table-set-index
 	  tae-translation-table-set-index)
-	 pr-el
+	 mes-el
 	 table-set table-syms table-syms-nested-p)
-    (while (setq pr-el (car products))
-      (setq products (cdr products))
-      (cond ((eq 'table-set (car pr-el))
-	     (setq table-set (cdr pr-el)
-		   table-syms
-		   (append table-syms
-			   (get-table-set-symbol-tables table-set))
+    (while (setq mes-el (car mes-cur))
+      (setq mes-cur (cdr mes-cur))
+      (cond ((eq 'table-set (car mes-el))
+	     (if table-set
+		 (error
+		  "Internal error. TAE message cannot have more than 1 table-sets:%S"
+		  messages))
+	     (setq table-set (cdr mes-el)
+		   table-syms (get-table-set-symbol-list table-set)
 		   result-myo (mucs-ccl-myo-add-table-set
 			       table-set result-myo))
 	     (if (table-set-nested-p table-set)
 		 (setq table-syms-nested-p t)))
 
-	    ((eq 'new-type (car pr-el))
-	     (setq mucs-current-type (cdr pr-el)))
+	    ((eq 'new-type (car mes-el))
+	     (setq mucs-current-type (cdr mes-el)))
 
 	    (t
-	     (error "Internal error.  unknown TAE message:%S" pr-el)
+	     (error "Internal error.  unknown TAE message:%S" mes-el)
 	     )))
+
     (if table-syms-nested-p
 	(mucs-ccl-myo-add-ccl
 	 `((r1 = 0)
@@ -1090,7 +1104,7 @@ NAME must be a declared translation."
 	      funcs (cdr funcs))
 	(cond ((table-set-p elem)
 	       (setq ctables
-		     (get-table-set-symbol-tables elem decodep)
+		     (get-table-set-symbol-list elem decodep)
 		     tables (append tables ctables)))
 	       (t
 		(error "Not yet supported type! %S" elem))))
